from .base import BaseAPI


class TMY(BaseAPI):
    ENDPOINT = "tmy"

    def __init__(self, lat, lon, **kwargs):
        """
        Typical meteorological year.

        Obligatory parameters:
        :param lat: Latitude in decimal degrees (south is negative).
        :param lon: Longitude in decimal degrees (west is negative).

        Describes the various output variables from the API call:

        - `G(h)`: Global irradiance on the horizontal plane (units: W/m2).
        - `Gb(n)`: Beam/direct irradiance on a plane always normal to sun rays (units: W/m2).
        - `Gd(h)`: Diffuse irradiance on the horizontal plane (units: W/m2).
        - `IR(h)`: Surface infrared (thermal) irradiance on a horizontal plane (units: W/m2).
        - `RH`: Relative humidity (units: %).
        - `SP`: Surface (air) pressure (units: Pa).
        - `T2m`: 2-m air temperature (units: degree Celsius).
        - `WD10m`: 10-m wind direction (0 = N, 90 = E) (units: degree).
        - `WS10m`: 10-m total wind speed (units: m/s).
        """
        super().__init__(lat, lon, **kwargs)

    def _get_endpoint(self):
        """
        Returns the endpoint URL for the Daily Radiation API call.
        """
        return self.BASE_URL + self.ENDPOINT

    def set_params(self, **kwargs):
        """
        Update or set parameters.
        """
        self.params.update(kwargs)

    @property
    def params(self):
        """
        Construct and return parameters for the API call.
        """
        parameters = {
            "lat": self.lat,
            "lon": self.lon,
            "usehorizon": self._params.get("usehorizon", 1),
            "startyear": self._params.get("startyear", None),
            "endyear": self._params.get("endyear", None),
            "outputformat": self._params.get("outputformat", 0),
            "browser": self._params.get("browser", 0)
        }

        # Remove any parameters set to None
        return {k: v for k, v in parameters.items() if v is not None}

    def months_selected(self):
        if self.data is None:
            self.fetch_data()

        return self.data["outputs"]["months_selected"]

    def hourly(self):
        """
        Hourly data over a tmy.

        :return: dict of API return "tmy_hourly"
        Keys:
        - `G(h)`: Global irradiance on the horizontal plane (units: W/m2).
        - `Gb(n)`: Beam/direct irradiance on a plane always normal to sun rays (units: W/m2).
        - `Gd(h)`: Diffuse irradiance on the horizontal plane (units: W/m2).
        - `IR(h)`: Surface infrared (thermal) irradiance on a horizontal plane (units: W/m2).
        - `RH`: Relative humidity (units: %).
        - `SP`: Surface (air) pressure (units: Pa).
        - `T2m`: 2-m air temperature (units: degree Celsius).
        - `WD10m`: 10-m wind direction (0 = N, 90 = E) (units: degree).
        - `WS10m`: 10-m total wind speed (units: m/s).
        - `time(UTC)`: timestamp
        """
        if self.data is None:
            self.fetch_data()
        return self.data["outputs"]["tmy_hourly"]

    def yearly_irradiation(self, irradiance_type: str = "global"):
        """
        Returns the total yearly irradiance.

        :param irradiance_type: Type of irradiance to compute. Choices are 'global', 'direct', and 'diffuse'.
                                'global' is the sum of 'direct' and 'diffuse'.
        :return: irradiance [W/m2] on the horizontal plane (global / diffuse) or always normal to sun rays (direct).
        """

        if self.data is None:
            self.fetch_data()

        total_irradiance = 0
        for hour in self.data["outputs"]["tmy_hourly"]:
            # Make sure that param global=1, else the api doesnt return irradiance.
            if irradiance_type == "global":
                total_irradiance += hour.get("G(h)", 0)
            elif irradiance_type == "direct":
                total_irradiance += hour.get("Gb(n)", 0)
            elif irradiance_type == "diffuse":
                total_irradiance += hour.get("Gd(h)", 0)
            else:
                raise ValueError("Invalid irradiance_type. Choose from 'global', 'direct', or 'diffuse'.")

        return total_irradiance



import json
import requests


class BaseAPI:
    BASE_URL = "https://re.jrc.ec.europa.eu/api/v5_2/"
    BASE_URL_V1 = "https://re.jrc.ec.europa.eu/api/v5_1/"

    def __init__(self, lat: float, lon: float,  **kwargs):
        """
        Constructor to initialize any common parameters for API calls.
        """
        if -90 <= lat <= 90 and -180 <= lon <= 180:
            self.lat = lat
            self.lon = lon
        else:
            raise ValueError("Incorrect Latitude or Longitude.")

        self._params = kwargs
        self.data = None

    @property
    def params(self):
        print(self._params)
        return self._params

    @params.setter
    def params(self, value):
        self._params = value

    def _get_endpoint(self):
        """
        Returns the endpoint URL for the specific API. 
        This should be overridden by subclasses.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def fetch_data(self):
        """
        Fetch data from the API.
        """
        print("---> FETCHING...")
        response = requests.get(self._get_endpoint(), params=self.params)
        self.data = self._handle_response(response)

    def _handle_response(self, response):
        """
        Handle the API response. Can be extended to handle errors, 
        convert formats, etc.
        """
        if response.status_code != 200:
            # Handle error
            self._handle_error(response)

        # Always use json, only let user decide for export
        if self._params["outputformat"] == "csv":
            return response.text
        elif self._params["outputformat"] == "json":
            return response.json()
        elif self._params["outputformat"] == "basic":
            return response.json()
        else:
            raise Exception(f"Invalid Outputformat.")

    def _handle_error(self, response):
        """
        Handle errors from the API response.
        This can be extended to raise custom exceptions based on the error type.
        """
        raise Exception(f"API Error {response.status_code}: {response.text}")

    def export(self, filename):
        file = open(filename, "w")
        json.dump(self.data, file)



from .base import BaseAPI


class Hourly(BaseAPI):
    ENDPOINT = "seriescalc"

    def __init__(self, lat, lon, pvcalculation: bool, angle: float = 0, aspect: float = 0, peakpower: float = None,
                 loss: float = None,
                 startyear: int = None, endyear: int = None, pvtech: str = "crystSi", **kwargs):
        """
        Hourly averages data.

        :param lat: Latitude in decimal degrees (south is negative).
        :param lon: Longitude in decimal degrees (west is negative).
        :param pvcalculation: If False outputs only solar radiation calculations.
                              If True outputs the estimation of hourly PV production as well.
        :param peakpower: Nominal power of the PV system, in kW.
        :param loss: Sum of system losses, in percent. E.g. 20% -> loss=20
        :param pvtech: (Default: 'crystSi') PV technology.{"crystSi", "CIS", "CdTe", "Unknown"}
        :param angle: (Default: 0) Inclination angle from horizontal plane. Not relevant for 2-axis tracking.
        :param aspect: (Default: 0) Orientation (azimuth) angle of the (fixed) plane.
                       0=south, 90=west, -90=east. Not relevant for tracking planes.

        Describes the various output variables from the API call:

        - `G(i)`: Global irradiance on the inclined plane (plane of the array) (units: W/m2).
        - `H_sun`: Sun height (units: degree).
        - `Int`: Indicates solar radiation values; 1 means values are reconstructed.
        - `P`: PV system power (units: W).
        - `T2m`: 2-m air temperature (units: degree Celsius).
        - `WS10m`: 10-m total wind speed (units: m/s).
        """
        self.pvcalculation = pvcalculation

        if endyear < startyear:
            raise ValueError("Incorrect time period. The calculation period for this app should be at least 1 years.")
        if startyear not in range(2005, 2020):
            raise ValueError("Incorrect start year. Please, enter an integer between 2005 and 2016.")
        self.startyear = startyear
        if endyear not in range(2005, 2020):
            raise ValueError("Incorrect end year. Please, enter an integer between 2005 and 2016.")
        self.endyear = endyear

        if pvcalculation and peakpower is None:
            raise ValueError("Invalid peak-power. If pvcalculation set True, peakpower can't be None.")
        self.peakpower = peakpower

        if pvcalculation and loss is None:
            raise ValueError("Invalid loss. If pvcalculation set True, loss can't be None.")
        if int(loss) not in range(0, 100):
            raise ValueError("Invalid loss value. Please, enter a float between 0 and 100.")
        self.loss = loss

        if pvtech not in ["crystSi", "CIS", "CdTe", "Unknown"]:
            raise ValueError("Invalid PV Technology. Valid technologies are 'crystSi', 'CIS', 'CdTe', 'Unknown'")
        self.pvtech = pvtech

        if int(angle) not in range(0, 91):
            raise ValueError("Invalid angle. Please, enter a float between 0 and 90.")
        self.angle = angle

        if int(aspect) not in range(-180, 181):
            raise ValueError("Invalid aspect / azimuth. Please, enter a float between -180 and 180.")
        self.aspect = aspect

        super().__init__(lat, lon, **kwargs)

    def _get_endpoint(self):
        """
        Returns the endpoint URL for the Daily Radiation API call.
        """
        return self.BASE_URL + self.ENDPOINT

    def set_params(self, **kwargs):
        """
        Update or set parameters.
        """
        self.params.update(kwargs)

    @property
    def params(self):
        """
        Construct parameters for the API call.
        """
        parameters = {
            "lat": self.lat,
            "lon": self.lon,
            "usehorizon": self._params.get("usehorizon", 1),
            "raddatabase": self._params.get("raddatabase", "PVGIS-SARAH2"),
            "startyear": self.startyear,
            "endyear": self.endyear,
            "pvcalculation": 1 if self.pvcalculation else 0,
            "peakpower": self.peakpower,
            "pvtechchoice": self.pvtech,
            "mountingplace": self._params.get("mountingplace", None),
            "loss": self.loss,
            "trackingtype": self._params.get("trackingtype", None),
            "angle": self.angle,
            "aspect": self.aspect,
            "optimalinclination": self._params.get("optimalinclination", None),
            "optimalangles": self._params.get("optimalangles", None),
            "components": self._params.get("components", None),
            "outputformat": self._params.get("outputformat", 0),
            "browser": self._params.get("browser", 0)
        }

        # Remove any parameters set to None
        return {k: v for k, v in parameters.items() if v is not None}

    def fetch_data(self):
        """
        Fetch data and change start and endyear if set to None before.
        """
        super().fetch_data()

        self.startyear = self.data["inputs"]["meteo_data"]["year_min"]
        self.endyear = self.data["inputs"]["meteo_data"]["year_max"]

    def hourly(self):
        if self.data is None:
            self.fetch_data()

        hourly = self.data["outputs"]["hourly"]

        return hourly

    def yearly_pv_production(self):
        """
        Calculates the total pv power production in W over a year.

        :return: p in W
        """
        if self.data is None:
            self.fetch_data()

        hourly = self.data["outputs"]["hourly"]

        p = 0
        for hour in hourly:
            p += hour.get("P", 0)

        return p

    def monthly_pv_production(self):
        """
        Calculates the total pv power production in W over a month for a year.

        :return: p in W
        """
        if self.data is None:
            self.fetch_data()

        hourly = self.data["outputs"]["hourly"]

        p = 0
        for hour in hourly:
            p += hour.get("P", 0)

        return p



from .base import BaseAPI


class Monthly(BaseAPI):
    ENDPOINT = "tmy"

    def __init__(self, lat, lon, **kwargs):
        """
        Typical meteorological year.

        Obligatory parameters:
        :param lat: Latitude in decimal degrees (south is negative).
        :param lon: Longitude in decimal degrees (west is negative).

        Describes the various output variables from the API call:

        - `G(h)`: Global irradiance on the horizontal plane (units: W/m2).
        - `Gb(n)`: Beam/direct irradiance on a plane always normal to sun rays (units: W/m2).
        - `Gd(h)`: Diffuse irradiance on the horizontal plane (units: W/m2).
        - `IR(h)`: Surface infrared (thermal) irradiance on a horizontal plane (units: W/m2).
        - `RH`: Relative humidity (units: %).
        - `SP`: Surface (air) pressure (units: Pa).
        - `T2m`: 2-m air temperature (units: degree Celsius).
        - `WD10m`: 10-m wind direction (0 = N, 90 = E) (units: degree).
        - `WS10m`: 10-m total wind speed (units: m/s).
        """
        super().__init__(lat, lon, **kwargs)

    def _get_endpoint(self):
        """
        Returns the endpoint URL for the Daily Radiation API call.
        """
        return self.BASE_URL + self.ENDPOINT

    def set_params(self, **kwargs):
        """
        Update or set parameters.
        """
        self.params.update(kwargs)

    @property
    def params(self):
        """
        Construct and return parameters for the API call.
        """
        parameters = {
            "lat": self.lat,
            "lon": self.lon,
            "usehorizon": self._params.get("usehorizon", 1),
            "startyear": self._params.get("startyear", None),
            "endyear": self._params.get("endyear", None),
            "outputformat": self._params.get("outputformat", 0),
            "browser": self._params.get("browser", 0)
        }

        # Remove any parameters set to None
        return {k: v for k, v in parameters.items() if v is not None}


from .base import BaseAPI


class Daily(BaseAPI):
    ENDPOINT = "DRcalc"

    def __init__(self, lat, lon, month, **kwargs):
        """
        Daily radiation for one day in a specific month.
        Calculated in a TMY.

        Obligatory parameters:
        :param lat: Latitude in decimal degrees (south is negative).
        :param lon: Longitude in decimal degrees (west is negative).
        :param month: The number of the month, starting at 1 for January. A value of 0 indicates data for all months.


        Optional parameters:
        :param usehorizon: (Optional) Calculate considering shadows from a high horizon. Default is 1 for "yes".
        :param userhorizon: (Optional) Height of the horizon at equidistant directions around the point of interest, in degrees. Starting at north and moving clockwise. The series '0,10,20,30,40,15,25,5' would mean the horizon height is 0° due north, 10° for north-east, 20° for east, 30° for south-east, etc.
        :param raddatabase
        :param angle
        :param aspect
        :param global
        :param glob_2axis
        :param clearsky
        :param clearsky_2axis
        :param showtemperatures
        :param localtime
        :param outputformat
        :param browser

        API return values:
        :var G(i): [W/m2] Global irradiance on a fixed plane.
        :var Gb(i): [W/m2] Direct irradiance on a fixed plane.
        :var Gd(i): [W/m2] Diffuse irradiance on a fixed plane.
        :var time: [xx:00] Time in hours. Starting at 00:00.
        """
        self.month = month
        super().__init__(lat, lon, **kwargs)

    def _get_endpoint(self):
        """
        Returns the endpoint URL for the Daily Radiation API call.
        """
        return self.BASE_URL + self.ENDPOINT

    def set_params(self, **kwargs):
        """
        Update or set parameters.
        """
        self.params.update(kwargs)

    @property
    def params(self):
        """
        Construct and return parameters for the API call.
        """
        parameters = {
            "lat": self.lat,
            "lon": self.lon,
            "month": self.month,
            "usehorizon": self._params.get("usehorizon", 1),
            "raddatabase": self._params.get("raddatabase", "PVGIS-SARAH2"),
            "angle": self._params.get("angle", 0),
            "aspect": self._params.get("aspect", 0),
            "global": self._params.get("global", 1),
            "glob_2axis": self._params.get("glob_2axis", 0),
            "clearsky": self._params.get("clearsky", 0),
            "clearsky_2axis": self._params.get("clearsky_2axis", 0),
            "showtemperatures": self._params.get("showtemperatures", 0),
            "localtime": self._params.get("localtime", 0),
            "outputformat": self._params.get("outputformat", 0),
            "browser": self._params.get("browser", 0)
        }

        # Remove any parameters set to None
        return {k: v for k, v in parameters.items() if v is not None}

    def total_irradiance(self, irradiance_type: str = "global"):
        """
        Returns the total irradiance during one day.

        :param irradiance_type: Type of irradiance to compute. Choices are 'global', 'direct', and 'diffuse'.
                                'global' is the sum of 'direct' and 'diffuse'.
        :return: irradiance [W/m2] on a fixed plane.
        """
        if self.data is None:
            self.fetch_data()

        total_irradiance = 0
        for hour in self.data["outputs"]["daily_profile"]:
            # Make sure that param global=1, else the api doesnt return irradiance.
            if irradiance_type == "global":
                total_irradiance += hour.get("G(i)", 0)
            elif irradiance_type == "direct":
                total_irradiance += hour.get("Gb(i)", 0)
            elif irradiance_type == "diffuse":
                total_irradiance += hour.get("Gd(i)", 0)
            else:
                raise ValueError("Invalid irradiance_type. Choose from 'global', 'direct', or 'diffuse'.")

        return total_irradiance

    def irradiance(self, as_list: bool = False):
        """
        Returns the total irradiance during one day. All types.
        """
        if self.data is None:
            self.fetch_data()

        if as_list:
            irradiance = {"G(i)": [], "Gb(i)": [], "Gd(i)": []}
        else:
            irradiance = {"G(i)": 0, "Gb(i)": 0, "Gd(i)": 0}

        for hour in self.data["outputs"]["daily_profile"]:
            # Make sure that param global=1, else the api doesnt return irradiance.
            if as_list:
                irradiance["G(i)"].append(hour.get("G(i)", 0))
                irradiance["Gb(i)"].append(hour.get("Gb(i)", 0))
                irradiance["Gd(i)"].append(hour.get("Gd(i)", 0))
            else:
                irradiance["G(i)"] += hour.get("G(i)", 0)
                irradiance["Gb(i)"] += hour.get("Gb(i)", 0)
                irradiance["Gd(i)"] += hour.get("Gd(i)", 0)

        return irradiance
